#lang sicp

;; Exercise 4.40.  Implementation of the multiple dwelling problem in
;; normal scheme.

;; Generate all possible assignment of people to floors then filter
;; them using the required predicates.

(define cooper-floors '(2 3 4 5))
(define baker-floors '(1 2 3 4))
(define fletcher-floors '(2 3 4))
(define miller-floors '(3 4 5))
(define smith-floors '(1 2 3 4 5))


;; From chapter 2

(define (filter predicate sequence)
  (cond ((null? sequence) '())
        ((predicate (car sequence))
         (cons (car sequence)
               (filter predicate (cdr sequence))))
        (else (filter predicate (cdr sequence)))))

(define (accumulate op initial sequence)
  (if (null? sequence)
      initial
      (op (car sequence)
          (accumulate op initial (cdr sequence)))))

(define (flatmap proc seq)
  (accumulate append '() (map proc seq)))

;; From chapter 4
(define (distinct? items)
      (cond ((null? items) true)
            ((null? (cdr items)) true)
            ((member (car items) (cdr items)) false)
            (else (distinct? (cdr items)))))


;; Enumerate all 5-tuples from the lists a b c d e
(define (quintuples a b c d e)
  (flatmap
   (lambda (i)
     (flatmap
      (lambda (j)
        (flatmap
         (lambda (k)
           (flatmap
            (lambda (l)
              (map (lambda (m)
                     (list i j k l m))
                   e))
            d))
         c))
      b))
   a))

(define (people-floors-ok? people)
  (let ((baker (list-ref people 0))
        (cooper (list-ref people 1))
        (fletcher (list-ref people 2))
        (miller (list-ref people 3))
        (smith (list-ref people 4)))
    (and (not (= (abs (- smith fletcher)) 1))
         (not (= (abs (- fletcher cooper)) 1))
         (distinct? (list baker cooper fletcher miller smith))
         (not (= baker 5))
         (not (= cooper 1))
         (not (= fletcher 5))
         (not (= fletcher 1))
         (> miller cooper))))


;; Enumerate all possible allocations of people to floors (removing
;; the obviously impossible ones), filtering using the predicate.
(define (multiple-dwelling)
  (filter people-floors-ok?
          (quintuples cooper-floors
                      baker-floors
                      fletcher-floors
                      miller-floors
                      smith-floors)))
